/*****************************************************************************

###FIELD###
Name:   AmtLines
Short:  Indicates the total amount of lines in a text object.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>The total amount of lines in a text object can be read from this field.  If
there are no lines in the object, a value of zero will be returned.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Background
Short:  String-based field for defining a background colour.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>If you set the Background field then the entire text area will be cleared
to the colour that you specify.  The colour must be in hexadecimal or CSV
format - for example to create a pure red colour, a setting of
&quot;#ff0000&quot; or &quot;255,0,0&quot; would be valid.</p>

<p>The background can be switched on and off using the BACKGROUND bit in the
Flags field.</p>

###END###

*****************************************************************************/

static ERROR SET_Background(struct Text *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->Background.Red, &Self->Background.Green, &Self->Background.Blue, &Self->Background.Alpha);
   Self->Flags |= TXF_BACKGROUND;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   BottomMargin
Short:  Defines the amount of whitespace to leave at the bottom of the text page.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Text is drawn as if it is on a page of paper and the
BottomMargin determines the amount of whitespace that is created at the bottom
of that page.  This is useful in creating a layout where the text is
positioned slightly away from the edge of the display area.</p>

<p>This value can be set as a fixed pixel coordinate only.</p>

###SEE ALSO###
Fields: LeftMargin, RightMargin, TopMargin

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   CharLimit
Short:  Limits the amount of characters allowed in a text object's string.
Type:   LONG
Status: Set

###DESCRIPTION###
<p>Set the CharLimit field if you wish to limit the amount of characters that
can appear in a text object's string.  The minimum possible value is 0 for
no characters.</p>

<p>The CharLimit field is most useful for restricting the amount of characters
that a user can enter in an editable text object.</p>

###END###

*****************************************************************************/

static ERROR SET_CharLimit(struct Text *Self, LONG *Value)
{
   if (*Value < 0) return(ERR_OutOfRange);

   Self->CharLimit = *Value;

   /*** Check that the existing string is within the character limit ***/


   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   CursorColour
Short:  The colour used for the text cursor.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour used for the text cursor may be defined through this field.
The colour must be in hexadecimal or CSV format - for example to create a
pure red colour, a setting of &quot;#ff0000&quot; or &quot;255,0,0&quot;
would be valid.</p>

###END###

****************************************************************************/

static ERROR SET_CursorColour(struct Text *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->CursorRGB.Red, &Self->CursorRGB.Green, &Self->CursorRGB.Blue, &Self->CursorRGB.Alpha);
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   CursorColumn
Short:  Indicates the current column position of the cursor.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>If you need to query the current position of the cursor within an editable
text object, read this field to obtain the column value.  You can also
write to this field if you wish to forcibly alter the current position of the
cursor.</p>

###SEE ALSO###
Field: CursorRow

###END###

****************************************************************************/

static ERROR SET_CursorColumn(struct Text *Self, LONG *Value)
{
   if (*Value >= 0) {
      Self->CursorColumn = *Value;
      Redraw(Self);
      return(ERR_Okay);
   }

   return(ERR_Failed);
}

/****************************************************************************

###FIELD###
Name:   CursorRow
Short:  Indicates the current line position of the cursor.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>If you need to query the current position of the cursor within an editable
text object, read this field to obtain the row value.  You can also
write to this field if you wish to forcibly alter the current position of the
cursor.</p>

###SEE ALSO###
Field: CursorColumn

###END###

****************************************************************************/

static ERROR SET_CursorRow(struct Text *Self, LONG *Value)
{
   if (*Value >= 0) {
      if (*Value < Self->AmtLines) Self->CursorRow = *Value;
      else Self->CursorRow = Self->AmtLines - 1;
      Redraw(Self);
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Defines the drawable area for text graphics.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a text object it should be contained by
a drawable area, such as a Render object.  If you do this correctly then
the text will detect the drawable area and set the Drawable field to the
correct object.  However, there may be occasions where you want to
initialise a text object to a non-graphical container.  If this is the
case then you must set the Drawable field to point to a valid drawable
area, such as a Render object.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Duration
Short:  Determines the duration of an active effect, measured in milliseconds.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>If you set the Duration field in conjunction with a text effect, the
effect will last only as long as the duration that you specify.  The
duration is measured in milliseconds.  Once the effect ends, normal activity
will resume, although the effect may restart if you have used the EFFECTLOOP
flag.</p>

###SEE ALSO###
Field: Effect, EffectDelay

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Effect
Short:  Enables animated text effects.
Prefix: EFX
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Special effects are supported in the text object through the Effect field.
This allows you to create text displays suitable for presentations and
catching the eye of the user.  The following effects are currently available:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Type</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">FADE</td><td>This effect uses alpha blending techniques to cross-fade the text into the display.</td></tr>
<tr><td width="10%">FLASH</td><td>This effect flashes the all of the text to white, then drops down to the colour pre-assigned to the text object's font.</td></tr>
<tr><td width="10%">TYPIST</td><td>This effect prints the text strings gradually, as if someone is typing it in real time.</td></tr>
</table>

<p>An effect starts either when the EffectDelay field decrements to zero, or
the first drawable frame is reached in an animation sequence (refer to the Frames
field).  The effect will last until either the last frame is displayed, or the
duration of the effect ends, as determined by the Duration field.</p>

<p>When using text effects, you must preset the strings in the text object
before initialisation.  Note that it is generally unwise to push the limits of
a text object while effects are active, as certain activity may clash with
effects when they are enabled.</p>

###SEE ALSO###
Field: Duration, EffectDelay

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   EffectDelay
Short:  Sets the delay for text effects, in milliseconds.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you have enabled an effect, you can set the EffectDelay field to delay the
start of the effect for a set number of milliseconds.  If this field is not set,
the effect will either start immediately or when the first drawable frame
is reached in an animation sequence.</p>

###SEE ALSO###
Field: Duration, Effect

###END###

*****************************************************************************/

/****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flags that affect object behaviour.
Type:   LONG/FLAGS
Prefix: TXF
Status: Read/Set

###DESCRIPTION###
<p>Special flags that apply to the text class are listed in the following
table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">AREASELECTED</td><td>This read-only flag is set automatically when the user has selected an area of text.</td></tr>
<tr><td width="10%">AUTOCLEAR</td><td>This flag will cause the text object to clear the entered text when the enter key is pressed.</td></tr>
<tr><td width="10%">BACKGROUND</td><td>This flag is automatically set if the Background field is written.  You can turn off the background colour at any time by disabling this flag.</td></tr>
<tr><td width="10%">COMMANDLINE</td><td>Provided for convenience, this flag can be set on initialisation to enable the following options: AUTOCLEAR, ENTERACTIVATE, HISTORY and EDIT.</td></tr>
<tr><td width="10%">DISABLED</td><td>This read-only flag is set if the object has been disabled through the Disable action.</td></tr>
<tr><td width="10%">DOUBLECLICK</td><td>If enabled, double clicking on any line will cause all child objects to be activated.</td></tr>
<tr><td width="10%">EDIT</td><td>By default, user text editing is turned off.  You need to enable it by setting the EDIT flag.</td></tr>
<tr><td width="10%">EFFECTLOOP</td><td>If special effects are in use, setting the EFFECTLOOP flag allows the effect to repeat itself continuously when in an animation loop.</td></tr>
<tr><td width="10%">ENTERACTIVATE</td><td>When this flag is set, any child objects that belong to the text object will be activated when the user hits the enter key.</td></tr>
<tr><td width="10%">FOCUSACTIVATE&nbsp;&nbsp;</td><td>If the text object loses the focus when this flag is set, the children of the text object will be activated.</td></tr>
<tr><td width="10%">FORCECAPS</td><td>This flag forces the capitalisation of all text that is entered into the object.</td></tr>
<tr><td width="10%">GLOBALEDITING</td><td>Keeps the text object in edit mode regardless of where the user focus might be.</td></tr>
<tr><td width="10%">HISTORY</td><td>Enables historical browsing when the up/down cursor keys are used.  The number of historical lines stored can be set in the HistorySize field.</td></tr>
<tr><td width="10%">MULTISELECT</td><td>Allows the user to select/highlight multiple lines by holding the Ctrl key.</td></tr>
<tr><td width="10%">NOFXDRAW</td><td>When in special effect mode, redrawing normally takes place automatically.  You can turn off this behaviour by setting this flag.</td></tr>
<tr><td width="10%">OVERWRITE</td><td>This flag forces overwrite mode when the user enters information through the keyboard.  If the flag is not set, then insert mode is used.</td></tr>
<tr><td width="10%">SECRET</td><td>The SECRET flag should be used when the text that is input by the user should be hidden from the display.  This option is most commonly used for password entry.</td></tr>
<tr><td width="10%">SINGLESELECT</td><td>Enables text selection (with or without edit mode).  If in line mode, a single click will allow the user to highlight any line.</td></tr>
<tr><td width="10%">SYSKEYS</td><td>This flag is set by default to enable system key support.  System-keys provide support for common key combinations such as Ctrl-C, Ctrl-X etc.</td></tr>
<tr><td width="10%">TABENTRY</td><td>If set, pressing the tab key has the same effect as pressing the enter key.</td></tr>
<tr><td width="10%">WORDWRAP</td><td>If you want the words to wrap around when the text string reaches the width of its container, set this flag.</td></tr>
<tr><td width="10%">VARIABLE</td><td>This flag forces a text object to recalculate any variable references in its string data whenever a redraw occurs.  Such references are indicated through square brackets, e.g. [systemdisplay.width].  Refer to the StrTranslate() function in the Strings module for more information.</td></tr>
</table>

###END###

****************************************************************************/

static ERROR SET_Flags(struct Text *Self, LONG *Flags)
{
   LONG FontFlags;

   if (*Flags & TXF_WORDWRAP) {
      FontFlags = Self->Font->Flags | FTF_WORDWRAP;
      SetField(Self->Font, FID_Flags, FT_LONG, &FontFlags);
   }
   else if (Self->Font->Flags & FTF_WORDWRAP) {
      FontFlags = Self->Font->Flags & ~FTF_WORDWRAP;
      SetField(Self->Font, FID_Flags, FT_LONG, &FontFlags);
   }

   Self->Flags = *Flags;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Focus
Short:  Refers to the object that will be monitored for user focussing.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>By default, a text object will become active (i.e. capable of
receiving keyboard input) when its drawable container receives the focus.  If
you would like to change this so that a Text becomes active when some
other object receives the focus, refer to that object by writing its ID to
this field.</p>

###SEE ALSO###
Field: Input

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Font
Short:  Points to a Font object that controls the drawing of text.
Type:   struct Font *
Status: Read

###DESCRIPTION###
<p>To set the face, colour and other attributes of a text object's graphics,
you need to read the Font field and write your settings to the Font object
prior to initialisation.  For a list of all the fields that can be set, please
refer to the documentation for the Font class.</p>

###SEE ALSO###
Class: Font

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Forces a text object's graphic to be drawn to a specific frame.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to a valid frame number, the text graphic will only
be drawn when the frame of the container matches the Frame number in this
field.  When set to 0 (the default), the text graphic will be drawn
regardless of the container's frame number.</p>

###SEE ALSO###
Field: Frames

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frames
Short:  Determines a sequence of frames that a text graphic may be drawn to.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>For long animation sequences you can set the Frames field to determine
individual frames that a text graphic should be drawn to.  For instance,
if a drawable has been assigned 20 animation frames and you want a text
graphic to appear in frames 5, 6, 7, 8, 17 and 19, you would use this
frame specification:</p>

<pre>
   5-8,17,19
</pre>

<p>Individual frame sets are separated by any non-numeric character such as
a comma, and ranges can be specified through the '-' character.</p>

<p>Setting the Frames field will override any setting that you may have
previously set in the Frame field.</p>

###END###

*****************************************************************************/

static ERROR SET_Frames(struct Text *Self, STRING Value)
{
   STRING str;
   LONG frame, end, counter;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   /*** Calculate the size of the frame array that we are going to need ***/

   Self->TotalFrames = 0;
   str = Value;
   while (*str) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > Self->TotalFrames) Self->TotalFrames = frame;
      }
      else str++;
   }

   Self->TotalFrames++; /* Add an extra frame because we start from #1, not zero */

   /*** Allocate the frame array and record the sequence ***/

   if (AllocMemory((Self->TotalFrames+1) * sizeof(WORD), MEM_DATA, &Self->Frames, NULL) IS ERR_Okay) {
      str = Value;
      counter = 1;
      while (*str) {
         if ((*str >= '0') AND (*str <= '9')) {
            if ((frame = StrToInt(str)) > 0) {
               while ((*str >= '0') AND (*str <= '9')) str++;
               while ((*str) AND (*str <= 0x20)) str++;
               if (*str IS '-') {
                  while ((*str) AND ((*str < '0') OR (*str > '9'))) str++;
                  if ((end = StrToInt(str)) > 0) {
                     while (frame != end) {
                        Self->Frames[frame] = counter++;
                        if (frame < end) frame++;
                        else frame--;
                     }
                     Self->Frames[frame] = 1;
                  }
                  while ((*str >= '0') AND (*str <= '9')) str++;
               }
               else Self->Frames[frame] = 1;
            }
            else while ((*str >= '0') AND (*str <= '9')) str++;
         }
         else str++;
      }

      Self->Frames[Self->TotalFrames] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   Self->Frame = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a text object's drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A text object's drawing region can be given a fixed or relative height by
setting this field to the desired value.  To set a relative height, use the
FD_PERCENT flag when setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct Text *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerheight, ycoord, value;

   if (!(Self->Dimensions & DMF_VERTICAL_FLAGS)) {
      value = Self->Font->LineSpacing * Self->Font->LineCount;
   }
   else if (Self->Dimensions & DMF_FIXED_HEIGHT) value = Self->Height;
   else if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ObjectError(Self, ERH_GetHeight, ERR_GetField));

      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         value = Self->Height * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_Y) ycoord = Self->YCoord;
         else ycoord = containerheight * Self->YCoord / 100;

         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - ycoord - Self->YOffset;
         else value = containerheight - ycoord - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetHeight, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Text *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Text] A height of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Text] A height of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetHeight, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      struct Variable var;
      var.Type = FD_LARGE;
      GET_Height(Self, &var); Self->AreaHeight = var.Large;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for defining the colour to use when text is highlighted.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour that you set through the Highlight field will determine the
colour that is used in highlighting selected text.  If you do not set this
field, a default colour will be used for highlighting.</p>

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct Text *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->Highlight.Red, &Self->Highlight.Green, &Self->Highlight.Blue, &Self->Highlight.Alpha);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HScroll
Short:  If scrolling is required, use this field to refer to a horizontal scroll bar.
Type:   OBJECTID
Status: Read/Set

###DESCRIPTION###
<p>If you want to attach a horizontal scrollbar to a text object, set
this field to an object belonging to the Scroll class.  So long as the Scroll
object is set up to provide full scrollbar functionality, the user will be
able to scroll the text display along the horizontal axis.</p>

###SEE ALSO###
Field: VScroll
Class: Scroll

###END###

*****************************************************************************/

static ERROR SET_HScroll(struct Text *Self, OBJECTID *ScrollID)
{
   OBJECTPTR scroll;

   if (GetClassID(*ScrollID) IS ID_SCROLL) {
      if (AccessObject(*ScrollID, 3000, &scroll) IS ERR_Okay) {
         SetField(scroll, FID_Object, FT_LONG, &Self->Head.UniqueID);
         Self->HScroll = *ScrollID;
         Self->XPosition = 0;
         if (Self->Head.ObjectFlags & NF_INITIALISED) CalculateHScroll(Self);
         ReleaseObject(scroll);
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_ExclusiveDenied));
   }
   else {
      DPrintF("@SetField:","[Text] Attempt to set the HScroll field with an invalid object.");
      return(ERR_Failed);
   }
}

/*****************************************************************************

###FIELD###
Name:   Input
Short:  Refers to the Keyboard object that input should be read from.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>By default, all text objects will receive keyboard input from the
"SystemKeyboard" object (which is typically created during the boot sequence).
However, you may want the text object to receive its keyboard input from a
different source.  If that is the case, you can refer to that object by
pointing to it in this field.</p>

<p>The object that you specify must support data channels and be able to
output data in the Keypress format.  If the Input object does not meet
these requirements, the text object will not be able to accept information
from it.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   LeftMargin
Short:  Defines the amount of whitespace to leave at the left of the page.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Text is drawn as if it is on a page of paper and the
LeftMargin determines the amount of whitespace that is created on the left
of that page. This is useful in creating a layout where the text is
positioned slightly away from the edge of the display area.</p>

<p>This value can be set as a fixed pixel coordinate only.</p>

###SEE ALSO###
Fields: RightMargin, TopMargin, BottomMargin

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   LineLimit
Short:  Restricts the total number of lines allowed in a text object.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Set the LineLimit field if you would like to restrict the maximum number
of lines allowed in a text object.  It is quite common to set this field to
a value of 1 for input boxes that have a limited amount of space available.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Identifies the location of a text file to load.
Type:     STRING
Status:   Get/Set

###DESCRIPTION###
<p>If the data for a text object needs to be read from a file, the easiest
way to load it into the object is to set the Location field and have the
object load the data for you.  The alternative is to load the data manually
and then set the String field with a data pointer, or use the DataChannel
action.</p>

<p>Note that if you set the Location field after initialisation, the
object will automatically clear its content and reload data from the
location that you specify.</p>

###SEE ALSO###
Field: String

###END###

*****************************************************************************/

static ERROR GET_Location(struct Text *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
}

static ERROR SET_Location(struct Text *Self, STRING Value)
{
   struct acRead read;
   OBJECTPTR file;
   LONG size, i;
   STRING line;

   DPrintF("~Set.Location:","[Text] %s", Value);

   if (Self->Location) { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;

         if (Self->Head.ObjectFlags & NF_INITIALISED) {
            if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
               FID_Flags|TLONG,      FL_READ,
               FID_Location|TSTRING, Self->Location,
               TAGEND) IS ERR_Okay) {

               Self->NoUpdate++;

               Action(AC_Clear, Self, NULL);

               if ((GetField(file, FID_Size, FT_LONG, &size) IS ERR_Okay) AND (size)) {
                  if (AllocMemory(size, MEM_STRING|MEM_NOCLEAR, (APTR *)&line, NULL) IS ERR_Okay) {
                     read.Buffer = line;
                     read.Length = size;
                     if (Action(AC_Read, file, &read) IS ERR_Okay) {
                        ActionTags(AC_DataChannel, Self, Self->Head.UniqueID, DATA_TEXT,
                           1, line, read.Result, 1);
                     }
                     FreeMemory(line);
                  }
                  else {
                     Action(AC_Free, file, NULL);
                     Self->NoUpdate--;
                     return(StepObjectError(Self, ERH_SetField, ERR_AllocMemory));
                  }
               }
               else {
                  Action(AC_Free, file, NULL);
                  Self->NoUpdate--;
                  return(StepObjectError(Self, ERH_SetField, ERR_GetField));
               }

               Self->NoUpdate--;

               Redraw(Self);
               CalculateHScroll(Self);
               CalculateVScroll(Self);

               Action(AC_Free, file, NULL);
            }
            else return(StepObjectError(Self, ERH_SetField, ERR_CreateObject));
         }
      }
      else return(StepObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   StepBack();
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Origin
Short:  Similar to the Location field, but does not automatically load content if set.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>This field is identical to the Location field, with the exception that it
does not update the content of a text object if you set it after
initialisation.  Altering the Origin of a text object is useful if you want
to record a change to a file's location without causing a load operation.</p>

###SEE ALSO###
Field: Location

###END###

****************************************************************************/

static ERROR SET_Origin(struct Text *Self, STRING Value)
{
   LONG i;

   if (Self->Location) { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectionObject
Short:  If the text object should send input data to another object, set this field.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>When the user presses the enter key, you can send the entered data to
the field of another object.  In order to do this, you need to point to that
object through the SelectionObject field and specify the destination field name
by setting the SelectionField string.  So long as both fields are set correctly,
the destination object will always reflect the value currently entered
in the text object.</p>

###SEE ALSO###
Field: SelectionField

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   SelectionField
Short:  This field is used in conjunction with the SelectionObject field.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>This field needs to be set in conjunction with the SelectionObject field - please
refer to it for more detail.</p>

###SEE ALSO###
Field: SelectionObject

###END###

*****************************************************************************/

static ERROR SET_SelectionField(struct Text *Self, STRING Value)
{
   StrCopy(Value, Self->SelectionField, sizeof(Self->SelectionField));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   RightMargin
Short:  Defines the amount of whitespace to leave at the right side of the text page.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Text is drawn as if it is on a page of paper and the
RightMargin determines the amount of whitespace that is created at the right
of that page.  This is useful in creating a layout where the text is
positioned slightly away from the edge of the display area.</p>

<p>This value can be set as a fixed pixel coordinate only.</p>

###SEE ALSO###
Fields: LeftMargin, TopMargin, BottomMargin

###END###

*****************************************************************************/

/****************************************************************************

###FIELD###
Name:   SelectColumn
Short:  Indicates the column position of a selection's beginning.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>If the user has selected an area of text, the starting column of that
area will be indicated by this field.  If an area has not been selected, the
value of the SelectColumn field is undefined.</p>

<p>You can check whether or not an area has been selected by testing the
AREASELECTED bit in the <a href="#Flags">Flags</a> field.</p>

###SEE ALSO###
Field: SelectRow

###END###

****************************************************************************/

/****************************************************************************

###FIELD###
Name:   SelectRow
Short:  Indicates the line position of a selection's beginning.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>If the user has selected an area of text, the starting row of that
area will be indicated by this field.  If an area has not been selected, the
value of the SelectRow field is undefined.</p>

<p>You can check whether or not an area has been selected by testing the
AREASELECTED bit in the <a href="#Flags">Flags</a> field.</p>

###SEE ALSO###
Field: SelectColumn

###END###

****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   String
Short:  Text information can be written directly to a text object through this field.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>You can set the data of a text object by passing a string to this field.
Updating a text object in this fashion causes it to analyse the
string information for return codes, so you can still split the string data
into lines.  Any data that is in the text object when you set this field will be
deleted automatically.  The graphics will also be redrawn and any attached
Scroll objects will be recalculated.</p>

###SEE ALSO###
Method: AddLine

###END###

*****************************************************************************/

static ERROR GET_String(struct Text *Self, STRING *Value)
{
   *Value = NULL;
   if (Self->AmtLines > 0) {
      *Value = Self->Array[0].String;
      return(ERR_Okay);
   }
   return(ERR_NoData);
}

static ERROR SET_String(struct Text *Self, STRING String)
{
   STRING line;

   /*** Turn off graphical updates ***/

   Self->NoUpdate++;

   /*** Clear all content ***/

   Action(AC_Clear, Self, NULL);

   /*** Add the string information ***/

   if (String) {
      line = String;
      while ((line) AND (*line)) {
         ActionTags(MTX_AddLine, Self, line, -1, StrLineLength(line));
         line = StrNextLine(line);
      }
   }

   /*** Update the entire text area ***/

   Self->NoUpdate--;

   /* In commandline mode, the cursor is placed at the end of the string when
   ** a change occurs.
   */

   if (Self->LineLimit IS 1) {
      if (Self->Array[0].String) Self->CursorColumn = Self->Array[0].Length;
   }

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      CalculateHScroll(Self);
      CalculateVScroll(Self);
      Redraw(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TabFocus
Short:  Allows the user to hit the tab key to focus on other GUI objects.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>If this field points to another GUI object, the user will be able to use
the tab key to move to that object when entering information into the text
object.  (Technically this causes the focus to be set to that object, and
the text object will thus lose the focus.)</p>

<p>When a series of objects are 'chained' via tab focussing, the user will
have an easier time moving between objects through use of the keyboard.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   TextHeight
Short:  Indicator for the pixel height of all lines in a text object.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The total height of all lines in a text object can be measured by reading
the TextHeight.  The returned value compensates for vertical and gutter
spacing.  Wordwrap will be taken into account if the WORDWRAP bit has been
set in the Flags field.</p>

###SEE ALSO###
Field: TextWidth

###END###

*****************************************************************************/

static ERROR GET_TextHeight(struct Text *Self, LONG *Value)
{
   LONG lines, pagewidth, row;

   if ((Self->Flags & TXF_WORDWRAP) AND (Self->AmtLines > 0) AND (Self->DrawableWidth > 0)) {
      lines = 0;
      pagewidth = Self->AreaWidth - Self->LeftMargin - Self->RightMargin;
      for (row=0; row < Self->AmtLines; row++) {
         if (Self->Array[row].PixelLength >= pagewidth) {
            lines += ((Self->Array[row].PixelLength + pagewidth) / pagewidth);
         }
         else lines++;
      }
   }
   else lines = Self->AmtLines;

   /*** In edit mode, there is always at least 1 active line (so that you can enter text) ***/

   if ((lines < 1) AND (Self->Flags & TXF_EDIT)) lines = 1;

   *Value = (Self->Font->MaxHeight + (Self->Font->LineSpacing * (lines-1)));

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TextWidth
Short:  Readable field that returns the maximum width of the text string in pixels.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>If you need to find out the pixel width of a text object then you
can read it from this field.  The width that is returned is that of the
longest line in the text object.  Note that the longest line does not
necesarily have the most characters, unless the font is of a fixed size (a
line of 30 'w' characters will typically be longer than a line of 50 'i'
characters because of the difference in character width).</p>

<p>The value includes the LeftMargin and RightMargin fields if you have set
them.</p>

###SEE ALSO###
Field: TextHeight

###END###

*****************************************************************************/

static ERROR GET_TextWidth(struct Text *Self, LONG *Value)
{
   LONG longest, pagewidth, row, width;

   longest = 0;
   if ((Self->Flags & TXF_WORDWRAP) AND (Self->DrawableWidth > 0)) {
      pagewidth = Self->AreaWidth - Self->LeftMargin - Self->RightMargin;

      if (Self->AmtLines < 50) {
         /* Calculate an accurate value for the text width by asking
         ** the font object to return the pixel width of each line
         ** when word-wrapping is taken into consideration.
         */

         for (row=0; row < Self->AmtLines; row++) {
            if (Self->Array[row].String) {
               /* NB: We would normally use CalcStringWidth(), but it does not
               ** work for our purposes as it doesn't pay consideration to
               ** word wrapping.
               */

               SetFields(Self->Font, FID_String|TSTRING,  Self->Array[row].String,
                                     FID_WrapWidth|TLONG, pagewidth,
                                     TAGEND);

               GetField(Self->Font, FID_Width, FT_LONG, &width);

               if (width > longest) longest = width;
            }
         }
      }
      else {
         /*** Calculate an approximate value for the text width ***/
         for (row=0; row < Self->AmtLines; row++) {
            if (Self->Array[row].PixelLength > pagewidth) {
               longest = pagewidth;
            }
            else if (Self->Array[row].PixelLength > longest) {
               longest = Self->Array[row].PixelLength;
            }
         }
      }
   }
   else {
      for (row=0; row < Self->AmtLines; row++) {
         if (Self->Array[row].PixelLength > longest) {
            longest = Self->Array[row].PixelLength;
         }
      }
   }

   *Value = longest + Self->LeftMargin + Self->RightMargin;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TopMargin
Short:  Defines the amount of whitespace to leave at the top of the text page.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Text is drawn as if it is on a page of paper and the
TopMargin determines the amount of whitespace that is created at the top
of that page.  This is useful in creating a layout where the text is
positioned slightly away from the edge of the display area.</p>

<p>This value can be set as a fixed pixel coordinate only.</p>

###SEE ALSO###
Fields: LeftMargin, RightMargin, BottomMargin

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    Visible
Short:   Indicates the visibility of a text object.
Type:    LONG/BOOLEAN
Status:  Read

###DESCRIPTION###
<p>If you need to know if a text object is visible or hidden, you can read
this field to find out.  A TRUE value is returned if the
object is visible and FALSE is returned if the object is invisible.  Note
that visibility is subject to the properties of the container that the text
object resides in.  For example, if a text object is visible but is contained
within a drawable object that is invisible, the end result is that both
objects are actually invisible.</p>

<p>Visiblity is directly affected by the Hide and Show actions if you
wish to change the visiblity of a text object.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   VScroll
Short:  If scrolling is required, use this field to refer to a vertical scroll bar.
Type:   OBJECTID
Status: Read/Set

###DESCRIPTION###
<p>If you want to attach a vertical scrollbar to a text object, set this
field to an object belonging to the Scroll class.  So long as the Scroll
object is set up to provide full scrollbar functionality, the user will be
able to scroll the text display along the vertical axis.</p>

###SEE ALSO###
Fields: HScroll
Class: Scroll

###END###

*****************************************************************************/

static ERROR SET_VScroll(struct Text *Self, OBJECTID *ScrollID)
{
   OBJECTPTR scroll;

   if (GetClassID(*ScrollID) IS ID_SCROLL) {
      if (AccessObject(*ScrollID, 3000, &scroll) IS ERR_Okay) {
         SetField(scroll, FID_Object, FT_LONG, &Self->Head.UniqueID);
         Self->VScroll = *ScrollID;
         Self->YPosition = 0;
         if (Self->Head.ObjectFlags & NF_INITIALISED) CalculateVScroll(Self);
         ReleaseObject(scroll);
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_ExclusiveDenied));
   }
   else {
      DPrintF("@SetField:","[Text] Attempt to set the VScroll field with an invalid object.");
      return(ERR_Failed);
   }
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a text object's drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A text object's drawing region can be given a fixed or relative width by
setting this field to the desired value.  To set a relative width, use the
FD_PERCENT flag when setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct Text *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, xcoord, value;
   LONG width;

   if (!(Self->Dimensions & DMF_HORIZONTAL_FLAGS)) {
      GET_TextWidth(Self, &width);
      value = width;
   }
   else if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ObjectError(Self, ERH_GetWidth, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_WIDTH) value = Self->Width;
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         value = Self->Width * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_X) xcoord = Self->XCoord;
         else xcoord = containerwidth * Self->XCoord / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - xcoord - Self->XOffset;
         else value = containerwidth - xcoord - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetWidth, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct Text *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Text] A width of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Text] A width of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetWidth, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      struct Variable var;
      var.Type = FD_LARGE;
      GET_Width(Self, &var); Self->AreaWidth = var.Large;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Defines the horizontal position of the drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>This field defines the horizontal position of the text drawing area from
the left of its container.  The value can be set as a fixed pixel coordinate,
or as a percentage if the FD_PERCENT field is used.  The text string will be
clipped to fit inside this area, so if you were to set the TextX field to a
negative value then the initial set of characters would be clipped until the
left-hand border was reached.</p>

###SEE ALSO###
Fields: TextX, XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Text *Self, struct Variable *Value)
{
   struct RenderInfo info;
   DOUBLE width, value;

   if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
   else if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) IS ERR_Okay) {
      if (info.Width <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * info.Width / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = info.Width * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = info.Width - width - Self->XOffset;
         else value = info.Width - width - (info.Width * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / info.Width;
   }
   else return(ObjectError(Self, ERH_GetX, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetX, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Text *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetX, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      struct Variable var;
      var.Type = FD_LARGE;
      GET_X(Self, &var); Self->AreaX = var.Large;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Defines the horziontal offset of the drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>You can offset the drawable region of a text graphic by setting this field.
A setting of 0 would offset the drawing region against the right edge, while
a setting of 10 would place it 10 pixels off from the right edge. Negative
values can move the drawing region outside of the container's boundaries.</p>

<p>This field may also be set as a percentage if the FD_PERCENT flag is used.
Use of a percentage will cause the position of the drawing region to be
calculated relative to the width of the container.</p>

<p>Writing to this field will always turn off any previous setting in the
XCoord field.</p>

###SEE ALSO###
Fields: XCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Text *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetXOffset, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Text *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetXOffset, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      struct Variable var;
      var.Type = FD_LARGE;
      GET_Width(Self, &var); Self->AreaWidth = var.Large;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Defines the vertical position of the drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>This field determines the vertical position of the text drawing region from
the top of its container.  The value can be set as a fixed pixel coordinate,
or as a percentage if the FD_PERCENT field is used.</p>

###SEE ALSO###
Fields: TextY, YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Text *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
   else if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetY, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Text *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetY, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      struct Variable var;
      var.Type = FD_LARGE;
      GET_Y(Self, &var); Self->AreaY = var.Large;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Defines the vertical offset of the drawable region.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>You can offset the drawable region of a text graphic by setting this field.
A setting of 0 would place the drawing region at the bottom edge, while a
setting of 10 would place it 10 pixels off from the bottom edge. Negative
values can move the region outside of the container's boundaries.</p>

<p>This field may also be set as a percentage if the FD_PERCENT flag is used.
Use of a percentage will cause the position of the text to be calculated
relative to the height of the container.</p>

<p>Writing to this field will always turn off any previous setting in the
YCoord field.</p>

###SEE ALSO###
Fields: YCoord, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Text *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 3000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetYOffset, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Text *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetYOffset, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      struct Variable var;
      var.Type = FD_LARGE;
      GET_Height(Self, &var); Self->AreaHeight = var.Large;
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

